import inspect

import importlib
from tbot import configs
from telegram import Bot, ChatAction


class BotApp(Bot):
    commands = {}
    default_commands = {}
    commands_path = configs.COMMAND_ROOT_PATH
    default_commands_path = 'tbot.commands'
    bot_name = "ABC"

    def __init__(self, bot_name=None):
        if bot_name:
            self.bot_name = bot_name
        super(BotApp, self).__init__(token=configs.TUTORIAL_BOT_TOKEN)

    def make_configs(self):
        print('{} configured'.format(self.name))

    def command(self, command_name=None, default=True):
        def decorator(f):  # TODO  handle the case with duplicate command names
            if inspect.getmodule(f).__name__ == self.default_commands_path:
                self.default_commands.update({command_name: f})
            else:
                self.commands.update({command_name: f})
            return f

        return decorator

    def _register_commands(self):
        importlib.import_module(self.commands_path)
        importlib.import_module(self.default_commands_path)
        print(self.commands.keys())
        print(self.default_commands.keys())

    def start(self):
        self._register_commands()

    def action_typing(self, chat_id):
        self.send_chat_action(chat_id=chat_id, action=ChatAction.TYPING)


def create_bot():
    bot = BotApp()
    bot.make_configs()
    return bot



from tbot.utils import get_bot

bot = get_bot()


@bot.command(command_name="docs")
def unknown_handler(chat_id, *args, **kwargs):
    message = "Please read the docs-->"
    bot.send_message(chat_id, message)


@bot.command(command_name="unknown")
def unknown_handler(chat_id, *args, **kwargs):
    message = "Sorry, invalid command: {}".format(kwargs.get('input_command'))
    bot.send_message(chat_id, message)



from .utils import get_bot


class BotUpdateHandler(object):
    bot = get_bot()

    def __init__(self, update):
        self.update = update
        self.run_commands()

    def get_commands(self):
        ex_functions = []
        if not self.message:
            raise NotImplementedError
        if self.location:
            ex_function = self.get_command_from_registered('location')
            ex_functions.append({'function': self.get_command_from_registered('location'),
                                 'args': (),
                                 'kwargs': self.location})
        if self.entities:
            command_list, _ = self.parse_text()
            for command_name in command_list:
                ex_functions.append({'function': self.get_command_from_registered(command_name),
                                     'args': (),
                                     'kwargs': {'input_command': command_name}})
        elif self.text:
            ex_functions.append({'function': self.get_command_from_registered('unknown'),
                                 'args': (),
                                 'kwargs': {'input_command': self.text}})
        return ex_functions

    def parse_text(self):
        command_list = []
        text_command_list = []
        for command in self.entities:
            command_text = self.text[command.get('offset'):command.get('offset') + command.get('length')]
            if "/" in command_text:
                command_list.append(command_text.strip('/'))
            else:
                text_command_list.append(command_text)
        return command_list, text_command_list

    def run_commands(self):
        function_list = self.get_commands()
        for function_item in function_list:
            f = function_item.get('function')
            args = function_item.get('args')
            kwargs = function_item.get('kwargs')

            if callable(f) and isinstance(f, object):
                f(self.chat_id, (), *args, **kwargs)
            else:
                self.throw_invalid_command(*args, **kwargs)

    def throw_invalid_command(self, *args, **kwargs):
        self.get_command_from_registered('unknown')(chat_id=self.chat_id, *args, **kwargs)

    def get_command_from_registered(self, name=None):
        ex_function = self.bot.commands.get(name)
        if not ex_function:
            ex_function = self.bot.default_commands.get(name)
        return ex_function

    @property
    def message(self):
        return self.update.get('message')

    @property
    def chat_id(self):
        return self.message.get('chat')['id']

    @property
    def entities(self):
        return self.message.get('entities')

    @property
    def location(self):
        return self.message.get('location')

    @property
    def text(self):
        return self.message.get('text')


from functools import wraps

from django.http import JsonResponse

from .exception import TelegramBotExceptionBase


def handle_api_errors(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except (TelegramBotExceptionBase,) as e:
            return JsonResponse(error_response(errors=e.serialize(), message=e.message), status=200)

    return wrapped


def successful_response(data=dict, message=""):
    return JsonResponse({
        "success": True,
        "message": message or "successful response",
        "data": data,
        "errors": {}
    })


def error_response(errors=dict, message=""):
    return {
        "success": False,
        "message": message or "error",
        "data": {},
        "errors": errors
    }



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from .views import BaseBotAPIView
from .activation import BotActivationView

urlpatterns = [
    path('handler/', csrf_exempt(BaseBotAPIView.as_view()), name="handler"),
    path('activate/', BotActivationView.as_view(), name="activate"),
]



import json

from django.http import HttpResponse, JsonResponse
from rest_framework.views import APIView

from tbot.handler import BotUpdateHandler
from tbot.serializers import UpdateSerializer


class BaseBotAPIView(APIView):
    serializer_class = UpdateSerializer
    update_handler = BotUpdateHandler

    def get(self, request, *args, **kwargs):
        return HttpResponse("OK!")

    def post(self, request, *args, **kwargs):
        t_data = json.loads(request.body)
        serializer = self.serializer_class(data=t_data)
        serializer.is_valid(raise_exception=True)
        self.update_handler(serializer.validated_data)
        return JsonResponse({"ok": "POST request processed"})



class TelegramBotExceptionBase(Exception):
    code = 1
    message = ""
    success = False

    def __init__(self, message=None):
        if message:
            self.message = message

    def serialize(self):
        return {
            "error_code": self.code,
            "error_message": self.message,
        }


class ConfigurationError(TelegramBotExceptionBase):
    code = 10
    message = "Configuration error."



from django.conf import settings
from .exception import ConfigurationError

# TODO: create configuration function to handle the errors during configuration process.

if settings.DJANGO_TBOT_CONFIGS:
    configs = settings.DJANGO_TBOT_CONFIGS
else:
    raise ConfigurationError

COMMAND_ROOT_PATH = configs.get("COMMAND_ROOT_PATH")
TUTORIAL_BOT_TOKEN = configs.get("TUTORIAL_BOT_TOKEN")
BOT_WEBHOOK_URL = configs.get("WEBHOOK_URL")
BOT_APP_PATH = configs.get("APP_PATH")

# TODO raise specific error
if not all((COMMAND_ROOT_PATH, TUTORIAL_BOT_TOKEN, BOT_WEBHOOK_URL, BOT_APP_PATH)):
    raise ConfigurationError



from rest_framework import serializers


class BaseSerializer(serializers.Serializer):

    def update(self, instance, validated_data):
        pass

    def create(self, validated_data):
        pass


class LocationSerializers(BaseSerializer):
    latitude = serializers.CharField(required=False)
    longitude = serializers.CharField(required=False)


class ChatSerializer(BaseSerializer):
    id = serializers.IntegerField(required=False)
    is_bot = serializers.BooleanField(required=False)
    first_name = serializers.CharField(required=False)
    last_name = serializers.CharField(required=False)
    username = serializers.CharField(required=False)
    type = serializers.CharField(required=False)


class SenderSerializer(BaseSerializer):
    id = serializers.IntegerField()
    is_bot = serializers.BooleanField()
    first_name = serializers.CharField()
    last_name = serializers.CharField()
    username = serializers.CharField()
    language = serializers.CharField()


class CommandSerializer(BaseSerializer):
    type = serializers.CharField()
    length = serializers.IntegerField()
    offset = serializers.IntegerField()


class MessageSerializer(BaseSerializer):
    message_id = serializers.IntegerField()
    # sender = SenderSerializer(source='from')
    chat = ChatSerializer(many=False)
    # date = serializers.DateTimeField(required=False)
    location = LocationSerializers(required=False)
    entities = CommandSerializer(many=True, required=False)
    text = serializers.CharField(required=False)


class UpdateSerializer(BaseSerializer):
    update_id = serializers.IntegerField()
    message = MessageSerializer()



import requests
from django.http import HttpResponse
from rest_framework.views import APIView
from tbot import configs
from tbot.utils import get_bot

bot = get_bot()


class BotActivationView(APIView):
    def get(self, request, *args, **kwargs):
        try:
            response = requests.get(
                f"{bot.base_url}/setWebhook?url={configs.BOT_WEBHOOK_URL}/bot/handler/")
            if response.status_code is not 200:
                # TODO handle specific errors
                return HttpResponse("There are some problems.")
            return HttpResponse("Activated")
        except Exception as e:
            return HttpResponse("There are some problems like this:  %s" % e)
